
/* Define this in the MAKEFILE, not here */
/* #define DEBUG */ 

#include "sunddk.h" /* #defines the crucial SUNSOFT_BUYBACK magic token */

/* This stuff comes from SUNWxwddk */
#define NEED_EVENTS
#include    "os.h"
#include    "sun.h"
#include    "suninput.h"
#include    "input.h"
#include    "mi.h"

#include <fcntl.h>

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <stropts.h>
#include <sys/conf.h>
#include <sys/stat.h>

#include <X11/Sunowconfig.h>


static char *version="@(#) ddxBOLTwacom.c 1.17@(#)";

/* Attempt at writing my own XINPUT extention module, so that you can
 * use my wacom Graphire2 USB driver.
 * 
 * Note that this module is SPECIFICALLY FOR Xsun.
 * You cannot compile it into XFree or vanilla X11R6, without significant
 * additions/wrappers for it, since there is a lack of standardization
 * for Xserver module entry points.
 *
 * Note also that you NEED THE SUN X DDX, which can be freely downloaded
 * from soldc.sun.com somewheres.
 *
 * This does NOT allow you to use the graphire as your main mouse.
 * You also need a regular mouse.
 *
 * Top-level entry point is ddxBOLTwacomProc()
 * Then wacomGetEvents() and wacomEnqueue() do all the real work
 */

#define DO_PROXIMITY 1


/* This is the pressure value where pressure input from the 'pen' or
 * eraser is auto-translated into showing button1 'down' as well
 * It isnt strictly 'neccessary', since the pen itself
 * adds button1 when a certain pressure is reached...
 * 'cept that the pressure involved is around 75, which then
 * has sucked up 20% of the total 0-511 range.
 */

/*#define PRESSURE_THRESHOLD 5*/


/* Okay, kinda ugly. but I know that graphire tablets have
 * an event size of 8 bytes
 */
#define RAW_EVENT_SIZE 8

/* This holds "driver private" data for the tablet.
 * The server holds on to this for us, and passes it back to us via
 * a DevicePtr->devicePrivate pointer
 */
typedef struct {
	int fd;
	int lastx,lasty,lastpressure;
	unsigned char lastbutton;
	unsigned char proximity;

	/* Okay, really really yeukky hack, I know.
	 * But this lets us do the multiple-device hack.
	 * pen is 'primary', erasor is 'secondary'.
	 * See ddxBOLTwacomProc() internal comments.
	 */
	CARD8 primaryid, secondaryid,tertiaryid;
	int am_primary;
} tabletPriv;


/* Magic int values we have to use as device event types.
 * It is actually part of the XINPUT standard, that these vars
 * are externed from the Xserver.
 */
extern int	DeviceMotionNotify;
extern int	DeviceValuator;

extern int	DeviceButtonPress;
extern int	DeviceButtonRelease;
extern int	ProximityIn;
extern int	ProximityOut;


/*  wacomGetEvents().
 * Can be considered what the Sun X docs refer to as
 *       DeviceReadProc()
 *
 *the server has received a SIGPOLL for our device.
 *  (Although how it knows it is for OUR device, I'm not sure)
 * Pass back to the server a pointer to an array of firmware events,
 * and update the pointer-to-int to let it know how many.
 * NULL/0 are acceptible values.
 *
 * callAgain is set to true, if we return before reading all the
 *   data that is waiting for us on the device. In that case, the
 *   server will call us again, to finish the processing.
 *
 * In theory, the server will receive a signal for every new pen event.
 * However, it is unknown if we will see every signal. So we have to
 * be paranoid, I guess.
 * For now, just read one event at a time?
 * XXX later versions, read up to some max number, maybe.
 *
 * Apparently, the Xserver will not call us back until it has used up
 * all the events we have returned from a previous call.
 * So it is acceptible to use static event buffers here.
 *
 * This all seems very wierd. I think that we have a very quick, short
 * call here to more accurately associate timestamps with each event.
 * Then we do the actual conversion batch-style, later.
 *  (No wonder X pointer latency is often horrible)
 */
#define MAX_EV_READ 5
static XI_eventPtr
wacomGetEvents(DevicePtr X11devptr, int *numEvents, Bool *callAgain)
{
	struct timeval timenow;
	tabletPriv *privdata=(tabletPriv*)X11devptr->devicePrivate;
	static XI_event eventbuff[MAX_EV_READ];
	int readstat;
	static CARD8 rawevents[MAX_EV_READ][RAW_EVENT_SIZE];
	int maxbytes=MAX_EV_READ * RAW_EVENT_SIZE;
	int rawcount;

	if(privdata->am_primary != 1){
		/* Only the 'primary' instance should actually read
		 * the physical device
		 */

		*callAgain=FALSE;
		return NULL;
	}

	/* Doing an ioctl(NREAD) doesnt seem to work too well.
	 * So just mimic what sample sunTablet does: read as much
	 * as possible, since we have already set FNDELAY in opendev()
	 */
	readstat=read(privdata->fd,rawevents,maxbytes);
	if(readstat==maxbytes){
		*callAgain=TRUE;
	} else {
		*callAgain=FALSE;
	}
	*numEvents= readstat / RAW_EVENT_SIZE;

	gettimeofday(&timenow,NULL);

	for(rawcount=0 ; rawcount< *numEvents ; rawcount++){
		eventbuff[rawcount].time=timenow;
		eventbuff[rawcount].event=&rawevents[rawcount];

	}

	return (eventbuff);
}

/* figure out whether we need to send proximityOUT, or proximityIN, 
 * and set things as appropriate
 */
static void calculateProximityEvent(
	 unsigned char proxbyte,
	 BYTE *eventtype,BYTE *detail)
{
	*detail=0;
	if(proxbyte&0x80){
#ifdef DEBUG
		ErrorF("Sending ProximityIn\n");
#endif
		*eventtype=ProximityIn;
	} else {
#ifdef DEBUG
		ErrorF("Sending ProximityOut\n");
#endif
		*eventtype=ProximityOut;
	}
	
}

/* This is split out from wacomEnqueue, to keep that routine clean.
 * We've determined that we need to send a button event, rather than
 * just a plain old motion event. So now we need to figure out
 * what TYPE of event: ButtonPress vs ButtonRelease, and
 * which button.
 * We make the "bad" assumption that it is impossible for two buttons
 * to change state at the exact same time.
 * Note that "button1" is triggered by sufficient pressure on the pen.
 */
static void calculateButtonEvent(
	 unsigned char oldbuttons,
	 unsigned char newbuttons,
	 deviceKeyButtonPointer *eventptr)
{
	int flag=0;
	if((oldbuttons&0x1) != (newbuttons&0x1)){
		flag=newbuttons & 0x01;
		eventptr->detail=Button1;
	} else if((oldbuttons&0x2) != (newbuttons&0x2)){
		flag=newbuttons & 0x02;
		eventptr->detail=Button2;
	} else if((oldbuttons&0x4) != (newbuttons&0x4)){
		flag=newbuttons & 0x04;
		eventptr->detail=Button3;
	}
	if(flag >0){
#ifdef DEBUG
		ErrorF("ddxBOLTwacom doing buttonPRESS\n");
#endif
		eventptr->type=DeviceButtonPress;
	} else {
#ifdef DEBUG
		ErrorF("ddxBOLTwacom doing buttonRelease for button\n");
#endif
		eventptr->type=DeviceButtonRelease;
	}
}

/* We are passed ONE event, from the array that wacomGetEvents has
 * just finished handing off to the server.
 * Each 'event' has data in the format that wacomGetEvents deals with.
 * Now hand it BACK to the server, in the format that the server
 * expects.
 * Damn stupid architecture, but thats what I have got to work with.
 * Why the hell not just require wacomGetEvents to convert in the
 * first place...?
 * I think this is a latency issue, and for some reason, the server
 * coders decided that it was important to shove the timestamps
 * on the raw data first.
 */
static void
wacomEnqueue(DevicePtr X11devptr, XI_eventPtr rawev_ptr)
{
	tabletPriv *privdata=(tabletPriv*)X11devptr->devicePrivate;
	int x,y,p;
	int rootx,rooty;
	unsigned char newbutton,newproximity;
	CARD8 *devdata=rawev_ptr->event;
	CARD8 srcdevid;
	/* ugliness done to make alignment work out right, or something */
	xEvent eventalloc[3];
	deviceKeyButtonPointer *serverevent=(deviceKeyButtonPointer*)&eventalloc[0];
	deviceValuator *valuatorevent=(deviceValuator *)&eventalloc[1];

	srcdevid=privdata->primaryid; /* default */
	/**********************************************************
	 * Note that if we ever wanted to allow the tablet to be
	 * the "primary pointer", we would have to encode things
	 * differently. Screw that. Keep the code simple.
	 *
	 * Except that XINPUT is far from simple. We need two stinking Xevents
	 * for one lousy action.
	 * First, an event saying, "The next event is a MotionEvent",
	 * then the actual motionevent.
	 ***********************************************************
	 */

	/***** Lots of hardcode assumptions for graphire2 tablet ****/
	x=(devdata[3]<<8)|devdata[2];
	y=(devdata[5]<<8)|devdata[4];
	p=(devdata[7]<<8)|devdata[6];
	switch(devdata[1] & 0x60){
		case 0x00: /* PEN */
			break;
		case 0x20: /* eraser */
			srcdevid=privdata->secondaryid;
			break;
		case 0x40: /* mouse */
			srcdevid=privdata->tertiaryid;
		default: /* ???. default to pen */
			break;
	}

	newbutton=devdata[1] & 0x0f;
#ifdef PRESSURE_THRESHOLD
	if(p  > PRESSURE_THRESHOLD){
		newbutton |= 0x01;
	}
#endif /* PRESSURE_THRESHOLD */


	/* This should really be called miGETPointerPosition */
	miPointerPosition(&rootx, &rooty);

	serverevent->time=TVTOMILLI(rawev_ptr->time);
	serverevent->root_x = rootx;
	serverevent->root_y = rooty;
	serverevent->deviceid=(int)(srcdevid |MORE_EVENTS);
	

	/* determine whether we need to do proximity event,
	 * or button, or plain old motion event
	 */
#ifdef DO_PROXIMITY
	newproximity=devdata[1] & 0x80;

	if(newproximity != (privdata->proximity)){
		calculateProximityEvent(devdata[1],
		                     &serverevent->type,&serverevent->detail);
#ifdef DEBUG
		ErrorF("  newP=%x, oldP=%x\n",newproximity,privdata->proximity);
#endif

	} else
#endif /* DO_PROXIMITY */
 	if(newbutton!=privdata->lastbutton){
		calculateButtonEvent(privdata->lastbutton,newbutton,
		                     serverevent);
		privdata->lastbutton=newbutton;
	} else {
		serverevent->type=DeviceMotionNotify;
		serverevent->detail = 0;
	}


#ifdef DO_PROXIMITY
	/* THis construct does not work! I'm bummed */
	/*if(privdata->proximity==newproximity==0){ */

	if((privdata->proximity==0) && (newproximity==0)){
		/* no new data to send! Out of range!*/
		return;
	} else {
		privdata->proximity = newproximity;
	}
#endif
#ifdef DEBUG
	ErrorF("x=%d, y=%d,p=%d: newbutton=%x\n",x,y,p,newbutton);
#endif

/**** Not sure whether to do this or not. 
 * does motionnotify normally hold button state??
 *****
	serverevent->state = 0;
	if(newbutton&0x01){
		serverevent->state |= Button1Mask;
	}
	if(newbutton&0x02){
		serverevent->state |= Button2Mask;
	}
	if(newbutton&0x04){
		serverevent->state |= Button3Mask;
	}
************/


	/* Since we provide 'absolute motion' data, we must
	 * always follow every single type of event we send,
	 * with a secondary 'DeviceValuator' type event
	 */

	valuatorevent->type=DeviceValuator;
	valuatorevent->deviceid=srcdevid;
	valuatorevent->device_state=0;
	valuatorevent->first_valuator=0;
	if(privdata->proximity==0){
		valuatorevent->num_valuators=0;
	} else {
		valuatorevent->num_valuators=3;
		valuatorevent->valuator0=x;
		valuatorevent->valuator1=y;
		valuatorevent->valuator2=p;
	}

	/*
	ErrorF("ddxBOLTwacom: enqueueing events of type %d,%d for devid %d\n",
	       serverevent->type,
	       valuatorevent->type,
		devintptr->id);
	*/

	/* Note: The server copies the data in serverevent */
	mieqEnqueue((xEventPtr)serverevent);
	mieqEnqueue((xEventPtr)valuatorevent);

}

/* We are required to have a function like this, to pass in to
 * InitValuatorClassDeviceStruct()
 * However, sincwe we do not record motion history, we dont have to
 * actually DO anything here.
 */
static int
wacomMotionProc(DeviceIntPtr ptr, xTimecoord *coords,
                unsigned long start, unsigned long stop,
                ScreenPtr s_ptr)
{
	return 0;
}


/* Open device.
 * Set mode of tablet.
 * Set "send-signal-on-io" mode
 * return 0 on FAIL, 1 on good, 2 on "device busy, assume secondary role"
 */
static int
opendev(DevicePtr X11devptr, char *devname)
{
	int newfd=-1;

	tabletPriv *privdata=(tabletPriv*)X11devptr->devicePrivate;

	if(privdata==NULL){
		Error("BOLTwacom:opendev: NULL privdata?");
		return (0);
	}

	newfd=open(devname, O_RDONLY);
	if(newfd <0){
		perror("BOLTwacom");
		ErrorF("BOLTwacom:opendev: could not open device\n");
		return (2);
	}
	privdata->fd=newfd;

	if (fcntl (privdata->fd, F_SETFL, (FNDELAY|FASYNC)) < 0) {
		perror("BOLTwacom");
		ErrorF("Can't set up tablet on fd %d\n", privdata->fd);
		return (0);

	}
	return (1);
}

/*
 * Register ourselves with server.
 * init server private stuff, and bind ourselves to handling user stuff.
 */
static int
initdev(DevicePtr X11devptr, tabletPriv *privdata,char *devname)
{
	DeviceIntPtr devintp=(DeviceIntPtr)X11devptr;
	Atom tablet_atom;
	int buttoncount=3; /* total of 3 buttons possible */
	/* a "map" of button codes (HUH?) */
	CARD8 buttonmap[4]={0,1,2,3}; 
	int num_axises, resolution;
	int max_x, max_y, max_pres;

	/* XXX Hardcoding to graphire-2 values that I have measured. */
	num_axises=3;
	resolution=1000;
	max_x=10208;
	max_y=7424;
	max_pres=511;


	/*******************************************************************
	 * Assume done with all the device driver opening, and             *
	 * internal status clearing. Now we can register with the X server *
	 *******************************************************************/

	tablet_atom=MakeAtom(devname,strlen(devname),FALSE);
	AssignTypeAndName(devintp, tablet_atom, devname);

	if(privdata->am_primary){
		/* Only register for IO signal handling,
		 * if we are the 'primary' device.
		 */

		if(RegisterFdIo(X11devptr, privdata->fd,
	                wacomGetEvents,wacomEnqueue) != Success)
		{
			ErrorF("BOLTwacom Failed on RegisterFdIo\n");
			return(!Success);
		}
	}

	InitButtonClassDeviceStruct(devintp, buttoncount, buttonmap);

	/*says, "expect 3 AxisStruct inits"*/
	InitValuatorClassDeviceStruct(devintp, num_axises,
					      wacomMotionProc,
					      0,
					      Absolute);

	/* last field, resolution, is in "counts per metre" according
	 * to X11R6 Input docs. But the sun docs are more vague,
	 * and the sample tablet code is wierder still.
	 */
	InitValuatorAxisStruct(devintp,0,0,max_x,max_x);
	InitValuatorAxisStruct(devintp,1,0,max_y,max_y);
	InitValuatorAxisStruct(devintp,2,0,max_pres,max_pres);

#ifdef DO_PROXIMITY
	/* a tablet reports "proximity" of the pen */
	InitProximityClassDeviceStruct(devintp);
#endif


	/* If we were interested in doing double-duty as the primary mouse,
	 * would also call RegisterXPointerInterest(), 
	 * RegisterSetDeviceValuatorsProc(), and
	 * InitFocusClassDeviceStruct()
	 */
	
	return (Success);
	
}

/* Ugh. I HATE THIS.
 * But I havent found another way to tell ddxBOLTwacomProc,
 * "This is the second time you have been called"
 * "This is the THIRD time you have been called"
 * Which is mandatory for being able to name the eraser and mouse
 * with unique XINPUT dev names.
 */

int staticcount=1;

/* BOLTwacomProc - bolthole.com top-level entry point for this
 *  wacom tablet ddx extension to Xsun.
 *
 * This can be considered what the Sun ddx docs refer to as
 *  "DeviceControlProc()"
 *
 * This routine gets called by the X server, by virtue of adding the
 * name of this function to OWConfig.
 *
 * Analogous to InitAndStartDevices(), in X11 XINPUT documentation. Sigh.
 *  The Xserver first calls us with devptr=??, mode=DEVICE_INIT.
 *
 */
int ddxBOLTwacomProc(DevicePtr X11devptr, int mode)
{
	tabletPriv *privdata;
	int numdevs=1;
	char *numdevstring;
	char *devstring="/dev/usb/wacom0";
	char devname[20];
	int devid;

	numdevstring=OWconfigGetAttribute("XINPUT","wacom","numdevs");
	if(numdevstring!=NULL){
		numdevs=atoi(numdevstring);
		if(numdevs<1) numdevs=1;
		OWconfigFreeAttribute(numdevstring);
	}
	/* if numdevs >1 in OWconfig, it means try for stub stuff, to
	 * fake second device as eraser
	 */

	if(X11devptr==NULL){
		ErrorF("ddxBOLTwacomProc: passed NULL DevicePtr\n");
		return (!Success);
	}

	privdata=(tabletPriv*)X11devptr->devicePrivate;
	

	switch(mode){

	    case DEVICE_INIT:

#ifdef DEBUG
		ErrorF("BOLTwacomProc: OWconf says numdevs=%d\n",numdevs);
#endif
		if(privdata != NULL)
		{
			/* only allowed to call init once */
			return (!Success);
		}

		/* We need to always set this on DEVICE_INIT.
		 * In preparation for a second call to this routine,
		 * with DEVICE_ON
		 */
		X11devptr->on=FALSE;

		/*****************************************
		 * This must be the first time we are called.
		 * set up for 'primary' id, which we assume
		 * will be for the 'pen'.
		 *
		 * Yes, this is a really, really creepy cast.
		 * But it somehow works. Based off the cast in
		 * the Enqueue() routine, which is actually based from
		 * Sun sample code.
		 */

		privdata=(tabletPriv*)xalloc(sizeof(tabletPriv));
		if(privdata==NULL) return (!Success);
		bzero(privdata,sizeof(tabletPriv));


		devid=((DeviceIntPtr)X11devptr)->id;
#ifdef DEBUG
		ErrorF("Debug: BOLTwacomProc called with devid= %d\n",
		       devid);
#endif

		X11devptr->devicePrivate=privdata;
		
		/* The devname should really be taken from the
		 * configs. arg.
		 */
		switch(opendev(X11devptr,devstring)){
			case 1:
				privdata->am_primary=1;
				privdata->primaryid=devid;
				if(numdevs>1)
					privdata->secondaryid=devid+1;
				if(numdevs>2)
					privdata->tertiaryid=devid+1;

#ifdef DEBUG
				ErrorF("wacom primary: opendev made fd=%d\n",
				       privdata->fd);
#endif
				break;
			case 2:
				if(numdevs<2){
					ErrorF(" BOLTwacom: cannot open device %s\n",
					       devstring);
					return(!Success);
				}
#ifdef DEBUG
				ErrorF(" Debug: BOLTwacom assuming we are stub\n");
#endif
				privdata->am_primary=0;
				break;
			default:
				ErrorF("opendev returned unexpected val\n");
				xfree(privdata);
				return(!Success);
				/* Assume if we can open the device, we should
				* take the primary role, and if we cant,
				* then someone else already has it
				*/
		}

		sprintf(devname,"wacomdev%d",staticcount++);

		if(privdata->am_primary==1){
			if(numdevs>1){
				privdata->secondaryid=privdata->primaryid+1;
#ifdef DEBUG
				ErrorF("BOLTwacomProc assuming secondary WILL BE devid %d\n",
				       privdata->secondaryid);
#endif
			}
			if(numdevs>2){
				privdata->tertiaryid=privdata->primaryid+2;
#ifdef DEBUG
				ErrorF("BOLTwacomProc assuming tertiary WILL BE devid %d\n",
				       privdata->tertiaryid);
#endif
			}
		}
		
#ifdef DEBUG
		ErrorF("wacom DEVICE_INIT calling initdev for %s\n",devname);
#endif
		initdev(X11devptr,privdata,devname);

		return (Success);

		break;

	    case DEVICE_ON:
		X11devptr->on = TRUE;

		if(privdata->am_primary==0){
#ifdef DEBUG
			ErrorF("wacom Secondary ignoring DEVICE_ON\n");
#endif
			break;
		}
		ErrorF("wacom Primary calling AddEnabledDevice for DEVICE_ON: fd=%d\n",
		       privdata->fd);
		AddEnabledDevice (privdata->fd);

		/* say "We want SIGPOLL" when new data arrives */

		ioctl(privdata->fd, I_SETSIG, S_INPUT);

#ifdef DEBUG
		ErrorF("wacom Primary done with ioctl I_SETSIG\n");
#endif

		
		break;
	    case DEVICE_OFF:
		if(privdata->am_primary){
			ioctl(privdata->fd, I_SETSIG, NULL);

			RemoveEnabledDevice (privdata->fd);
		}

		X11devptr->on = FALSE;
		break;

	    case DEVICE_CLOSE:
		if((X11devptr->on) &&
		   (privdata->am_primary))
		{
#ifdef DEBUG
			ErrorF("wacom Primary calling RemoveEnabledDevice\n");
#endif
			RemoveEnabledDevice (privdata->fd);
			close(privdata->fd);
		}
		xfree(X11devptr->devicePrivate);
		X11devptr->on = FALSE;
		break;
	}

	return (Success);
}


